
#pragma once

#include <tee/component.h>
#include <root/component.h>

namespace Tee {
	using Genode::Allocator;
	using Genode::Arg_string;
	using Genode::Env;
	using Genode::Insufficient_ram_quota;
	using Genode::error;
	using Genode::max;
	using Genode::size_t;

	template <class SESSION_COMPONENT> class Root;
};


template <class SESSION_COMPONENT>
class Tee::Root : public Genode::Root_component<SESSION_COMPONENT,
                                                Genode::Single_client>
{
	private:

		Env       &_env;
		Allocator &_md_alloc;

	protected:

		SESSION_COMPONENT *_create_session(const char *args) override
		{
			const size_t ram_quota   = Arg_string::find_arg(args, "ram_quota"  ).ulong_value(0);
			const size_t tx_buf_size = Arg_string::find_arg(args, "tx_buf_size").ulong_value(0);
			const size_t rx_buf_size = Arg_string::find_arg(args, "rx_buf_size").ulong_value(0);

			/* deplete ram quota by the memory needed for the session structure */
			const size_t session_size = max(4096UL, static_cast<size_t>(sizeof(SESSION_COMPONENT)));
			if (ram_quota < session_size)
				throw Genode::Insufficient_ram_quota();

			/*
			 * Check if donated ram quota suffices for both communication
			 * buffers and check for overflow
			 */
			if (tx_buf_size + rx_buf_size < tx_buf_size ||
			    tx_buf_size + rx_buf_size > ram_quota - session_size) {
				Genode::error("insufficient 'ram_quota', got ", ram_quota, ", "
				              "need ", tx_buf_size + rx_buf_size + session_size);
				throw Genode::Insufficient_ram_quota();
			}

			return new (Root::md_alloc())
			            SESSION_COMPONENT(tx_buf_size, rx_buf_size,
			                             _md_alloc, _env);
		}

	public:

		Root(Genode::Env &env, Genode::Allocator &md_alloc)
		: Genode::Root_component<SESSION_COMPONENT, Genode::Single_client>(&env.ep().rpc_ep(), &md_alloc),
			_env(env), _md_alloc(md_alloc)
		{ }
};

